"""
Init UniCeption Prediction Heads
"""

from uniception.models.prediction_heads.base import (
    AdaptorInput,
    AdaptorOutput,
    Covariance2DAdaptorOutput,
    MaskAdaptorOutput,
    PredictionHeadInput,
    PredictionHeadOutput,
    RegressionAdaptorOutput,
    RegressionWithConfidenceAdaptorOutput,
    RegressionWithConfidenceAndMaskAdaptorOutput,
    RegressionWithMaskAdaptorOutput,
    UniCeptionAdaptorBase,
    UniCeptionPredictionHeadBase,
)
